# --------------------------------------------------------------------------- #
# $Id: pllshared.py,v 1.16 2003/06/29 20:51:45 pandr Exp $
# --------------------------------------------------------------------------- #

# This is for stuff that is shared between pointless.py and pllbasics.py

import pllog
from string import *
import gettext
import re
from _pointless import LOCALEDIR
import types, exceptions

## HACK to work-around some problems in the old version of gettext.py
try:
	t = gettext.translation('pointless', LOCALEDIR)
	gt = t.gettext
except: 
	def gt(message):
		return (message)

# --------------------------------------------------------------------------- #

class EnumException(exceptions.Exception):
    pass

class Enumeration:
    def __init__(self, name, enumList):
        self.__doc__ = name
        lookup = { }
        reverseLookup = { }
        i = 0
        uniqueNames = [ ]
        uniqueValues = [ ]
        for x in enumList:
            if type(x) == types.TupleType:
                x, i = x
            if type(x) != types.StringType:
                raise EnumException, "enum name is not a string: " + x
            if type(i) != types.IntType:
                raise EnumException, "enum value is not an integer: " + i
            if x in uniqueNames:
                raise EnumException, "enum name is not unique: " + x
            if i in uniqueValues:
                raise EnumException, "enum value is not unique for " + x
            uniqueNames.append(x)
            uniqueValues.append(i)
            lookup[x] = i
            reverseLookup[i] = x
            i = i + 1
        self.lookup = lookup
        self.reverseLookup = reverseLookup
    def __getattr__(self, attr):
        if not self.lookup.has_key(attr):
            raise AttributeError
        return self.lookup[attr]
    def whatis(self, value):
		return self.reverseLookup[value]

eAdjustmentType = Enumeration("eAdjustmentType",
	[ "ADJUSTED",
	"FLUSH_LEFT",
	"FLUSH_RIGHT",
	"CENTERED"
	]
)

class v2:
	"""Two dimensional vector of floats"""
	def __init__(self, x, y):
		self.x = x
		self.y = y

class v3:
	"""Three dimensional vector of floats with overloaded operators"""
	def __init__(self, x, y, z):
		self.x = x
		self.y = y
		self.z = z
	def __add__(self, other):
		return v3(self.x + other.x, self.y + other.y, self.z + other.z)
	def __sub__(self, other):
		return v3(self.x - other.x, self.y - other.y, self.z - other.z)
	def __neg__(self):
		return v3(-self.x, -self.y, -self.z)
	def __str__(self):
		return "v3(%s,%s,%s)" % (self.x.__str__(), self.y.__str__(), 
			self.z.__str__())

class size:
	"""Two dimensional vector cast as range"""
	def __init__(self, min, max):
		self.min = min
		self.max = max
		assert(min<=max)
	def set(self, min, max):
		assert(min<=max)
		self.min, self.max = min, max
	def __str__(self):
		return "[%.3f, %.3f]" % (self.min, self.max)
	def size(self): return self.max - self.min

# --------------------------------------------------------------------------- #

class PllParseError(Exception):
	def __init__(self, args=None):
		self.args = args

class PllCommandError(Exception):
	def __init__(self, args=None):
		self.args = args

# --------------------------------------------------------------------------- #

class PlSize:
	ePixels = 1
	ePlSize = 2
	def __init__(self, value):
		if value[-2:] == 'px':
			self.value = atof(value[:-2])
			self.unit  =  PlSize.ePixels
		else:
			self.value = atof(value)
			self.unit  = PlSize.ePlSize
	def pixels(self, scale):
		if self.unit == PlSize.ePixels:
			return self.value
		elif self.unit == PlSize.ePlSize:
			return self.value * scale
	def plsize(self, scale):
		if self.unit == PlSize.ePixels:
			return self.value / scale
		elif self.unit == PlSize.ePlSize:
			return self.value

re_comment    = re.compile(r'^#.*')
re_command    = re.compile('^=([A-Za-z_][A-Za-z0-9_-]*)(\.[A-Za-z_][A-Za-z0-9_]*|)[\t ]*(\(|)')
re_begin      = re.compile('^begin-[A-Za-z_]')
re_linewrap   = re.compile(r'^\\\s*$')
re_escapes    = re.compile(r'^\\([=\\{} ])')
re_blockstart = re.compile(r'^{')
re_blockend   = re.compile(r'^}')
re_letters    = re.compile(r'^[^\s=]')
re_space      = re.compile(r'^\s')
re_newline    = re.compile(r'^\n')
re_eof        = re.compile(r'^$')

class Command:
	def __init__(self, name, execute=None, slidestep=0, transform=0, 
		break_condition=re_newline, vars = None):
		l = name.split(":")
		if len(l) == 2:
			self.name, self.argfmt = l
		else:
			self.name = name
			self.argfmt = ""
		self.exe = execute
		self.slidestep = slidestep
		self.transform  = transform
		self.break_condition = break_condition
		self.vars = vars
	def execute(self, args, parser, vars):
		assert(self.exe)
		if self.vars:
			return self.exe(args, parser, vars)
		else:
			return self.exe(args, parser)
	def parse_args(self, argstring):
		args = []
		largs = argstring.split(",")
		for fmt in self.argfmt:
			if not fmt:
				print gt('command ')+self.name+\
					_(' called with too many arguments\n')
				return []
			if len(largs) == 0:
				print gt('command ')+self.name+\
					_(' called with too few arguments\n')
				return []
			arg = (largs[0]).strip()
			largs = largs[1:] ## grr, an ugly default (just a single element) pop 
			if fmt == 'i':
				if arg[0:2] == "0x":
					args.append(atoi(arg,16))
				else:
					args.append(atoi(arg))
			elif fmt == 'f':
				args.append(atof(arg))
			elif fmt == 's':
				if arg[0] == '"' and arg[-1] == '"':
					args.append(arg[1:-1])
			elif fmt == 'S':
				args.append(PlSize(arg))
			elif fmt == 'c':
				if arg[0] == '"' and arg[-1] == '"':
					## color specification, e.g. "Black"
					color_table = {
						"black" : [0.0, 0.0, 0.0, 1.0],
   						"white" : [1.0, 1.0, 1.0, 1.0],
						"red" 	: [0.7, 0.0, 0.0, 1.0],
						"green" : [0.0, 0.7, 0.0, 1.0],
						"blue"	: [0.0, 0.0, 0.7, 1.0],
						"yellow": [1.0, 1.0, 0.3, 1.0],
						"cyan"	: [0.0, 0.7, 0.7, 1.0],
						"magenta" :  [0.7, 0.0, 0.7, 1.0],
						"dgray"	: [0.3, 0.3, 0.3, 1.0],
						"lgray"	: [0.7, 0.7, 0.7, 1.0],
						"lred"	: [1.0, 0.3, 0.3, 1.0],
						"lgreen" : [0.3, 1.0, 0.3, 1.0],
						"lblue"	: [0.3, 0.3, 1.0, 1.0],
						"brown"	: [0.7, 0.7, 0.0, 1.0],
						"lcyan"	: [0.3, 1.0, 1.0, 1.0],
						"lmagenta" : [1.0, 0.3, 1.0, 1.0]
					}
					s = arg[1:-1].lower()
					if s.isalpha():
						if color_table.has_key(s):
							args.extend(color_table[s])
						else:
							pllog.warn("Unable to find color %s in color_table" % s)
					else:
						pllog.warn("All characters in color %s must be alphabetic" % s)
				elif arg[0:2] == "0x" and ((len(arg) == 8) or (len(arg) == 10)):
					## color specification 0x0000ff or 0x0000ffff: , e.g. 0x05050F
					if len(arg) == 8:
						pllog.debug("No alpha specification")
						alpha = 255
					elif len(arg) == 10:
						alpha = 16*atoi(arg[8],16)+atoi(arg[9],16)
					args.extend( map (lambda x: (float)(x)/255, 
						[	16*atoi(arg[2],16)+atoi(arg[3],16), 
							16*atoi(arg[4],16)+atoi(arg[5],16), 
							16*atoi(arg[6],16)+atoi(arg[7],16), 
							alpha ] ) 
					)
				elif len(largs) >= 3:
					## color specification: ffff, e.g. 0.05, 0.5, 0.75, 1.0
					args.extend([
						atof(arg), 
						atof((largs[0]).strip()), 
						atof((largs[1]).strip()), 
						atof((largs[2]).strip())
					])
					largs = largs[2:] ## grr, lets pop off three more 
				else:
					pllog.warn(gt("Unable to handle color specification %s"\
						% arg[1:-1]))
					return []
			else:
				print gt('Unknown format value ')+fmt+"\n"
				return []
		return args
	def set(self, hash):
		for k in hash.keys():
			vs.set_value(self.name, k, hash[k])

class Variable:
	def __init__(self, name, localizable, value = None):
		self.name        = name
		self.localizable = localizable
		self.value       = value
	def __repr__(self): return "%s==%s" % (self.name, str(self.value))
	def set_value(self, value):
		self.value = value
	def get_value(self):
		return self.value
	def copy(self):
		return Variable(self.name, self.localizable, self.value)

class VariableStack:
	def __init__(self):
		self._variables = [{}]
	def get_variable(self, scope, variablename):
		for i in range(1, len(self._variables)+1):
			v = self._variables[-i]
			if v.has_key(scope) and v[scope].has_key(variablename):
				return v[scope][variablename]
		return None
	def push(self):
		self._variables.append({})
	def pop(self):
		self._variables.pop()
	def set_value(self, scope, variablename, value):
		# If var defined in current context, just set it
		if self.current().has_key(scope) and \
				self.current()[scope].has_key(variablename):
			self.current()[scope][variablename].set_value(value)
			return
		# Otherwise look for it further up
		var = self.get_variable(scope, variablename)
		if not var:
			var = self.new_var(scope, variablename, localizable = 1)
		if var.localizable:
			copy = var.copy()
			copy.set_value(value)
			if not self.current().has_key(scope):
				self.current()[scope] = {}
			self.current()[scope][variablename] = copy
		else:
			var.set_value(value)
	def get_value(self, scope, variablename):
		var = self.get_variable(scope, variablename)
		if var: return var.get_value()
		else:   return None
	def new_var(self, scope, varname, localizable):
		if not self.current().has_key(scope):
			self.current()[scope] = {}
		v = Variable(varname, localizable)
		self.current()[scope][varname] = v
		return v
	def current(self):
		return self._variables[-1]

vs = VariableStack()

class Scope:
	def __init__(self, scope, variablestack):
		self.scope = scope
		self.variablestack = variablestack
	def __getitem__(self, key):
		return self.variablestack.get_value(self.scope, key)
	def __setitem__(self, key, value):
		self.variablestack.set_value(key, scope, value)

# --------------------------------------------------------------------------- #

