# --------------------------------------------------------------------------- #
# $Id: pointless.py,v 1.309 2003/07/17 19:15:57 pandr Exp $
# --------------------------------------------------------------------------- #

# --------------------------------------------------------------------------- #
# Import stuff

import _pointless, sys
from _pointless import VERSION, LOCALEDIR
from random import random
from string import *
import getopt, sys, os, string, re
import ConfigParser
import weakref
import copy
import md5
from pllshared import *
from pllcontroller import *
import ploptions   # options and their default settings
import plsetup     # resource setup stuff
import pllog       # log module
import plfiles     # filehandler

# --------------------------------------------------------------------------- #
# Globals

__id__ = "$Id: pointless.py,v 1.309 2003/07/17 19:15:57 pandr Exp $"
renderer = None
pll = None
INFINITY = 1000
DEBUG = 0 ## flag used to turn on/off extra debugging

PROFILE = 0 ## flag used to turn on/off profiling
if PROFILE:
	import profile
	import pstats

# --------------------------------------------------------------------------- #
# Node and group objects

class Node:
	""" 
		DESCRIPTION:
			The base class for everything that can be in the scene graph.
			It is essentially the python wrapping of the Node class in the
			CPP world of pointless.
	"""
	def __init__(self, this, context):
		assert(self.__class__ != Node)
		self.controllers = []
		self.this        = None
		self._size       = v3(size(0,0), size(0,0), size(0,0))
		self._hrefs      = [v3(0,0,0)]
		self._vrefs      = [v3(0,0,0)]
		self.context     = context
		self.debug_draw  = 0
		self.discardable = 0
		if this: self.set_this(this)
	def __repr__(self):
		return "<"+self.name()+">"
	def set_this(self, this):
		self.this = this
		cppsize = _pointless.node_get_size(self.this)
		self._size.x.set(cppsize[0], cppsize[1])
		self._size.y.set(cppsize[2], cppsize[3])
		self._size.z.set(cppsize[4], cppsize[5])
	def name(self):
		if self.this: 
			return _pointless.node_name(self.this)
		else: 
			return "Node"
	def full_name(self):
		if self.this: 
			return _pointless.node_full_name(self.this)
		else: 
			return "Node"
	def get_pos(self):
		if self.this:
			pos = _pointless.node_get_pos(self.this)
			return v3(pos[0], pos[1], pos[2])
		else:
			return v3(0.0, 0.0, 0.0)
	def set_pos(self, pos):
		if self.this: 
			return _pointless.node_set_pos(self.this, int(pos.x), 
				int(pos.y), int(pos.z))
	def get_size(self): return self._size
	def set_visible(self, *args):
		if self.this: return _pointless.node_set_visible(self.this, *args)
	def set_alpha(self, *args):
		if self.this: return _pointless.node_set_alpha(self.this, *args)
	def _debug_draw(self): pass
	def init_debug_draw(self):
		if self.debug_draw: self._debug_draw()

class Leaf(Node):
	def __init__(self, this, context=None):
		Node.__init__(self, this, context)

class Stretchable(Node):
	""" 
		DESCRIPTION:
			The base class for all nodes that may change size by the process 
			of stretching (or shrinking) to fill available space. 
	"""
	def __init__(self, natural_width=0, stretchability=0, shrinkability=0,
			natural_height=0, context=None, stretch_dim=0):
		assert(self.__class__ != Stretchable) # must derive from this class
		Node.__init__(self, this=None, context=context)
		self.stretchability = stretchability
		self.shrinkability  = shrinkability
		self.set_natural(v3(natural_width, natural_height, 0))
		self.stretch_dim = stretch_dim
	def __repr__(self): return str(self._size)
	def set_natural(self, natural):
		self._natural = natural
		self._size = v3(size(0,natural.x), size(0, natural.y), \
			size(0, natural.z))
	def get_natural(self): return self._natural
	def set_stretch(self, stretch):
		self.set_width(self.get_natural().x + self.stretchability*stretch)
	def set_width(self, width):
		self.set_natural(v3(width, self.get_size().y.max, 0))

class Rectangle(Stretchable):
	""" 
		DESCRIPTION:
			This is essentially the python wrapping of the rectangle
			method in the nodefactory class in the CPP world of pointless.
	"""
	def __init__(self, natural, context, ll_color, lr_color=-1, ul_color=-1, 
			ur_color=-1):
		Stretchable.__init__(self, natural_width=natural.x, 
			stretchability=INFINITY*INFINITY, natural_height=natural.y, 
			context=context)
		self.ll_color = ll_color
		if lr_color == -1:
			self.lr_color = ll_color
		else:
			self.lr_color = lr_color
		if ul_color == -1:
			self.ul_color = ll_color
		else:
			self.ul_color = ul_color
		if ur_color == -1:
			self.ur_color = ll_color
		else:
			self.ur_color = ur_color
		x = _pointless.nodefactory_rectangle(natural.x, natural.y, 
			ll_color[0], ll_color[1], ll_color[2], ll_color[3],
			lr_color[0], lr_color[1], lr_color[2], lr_color[3],
			ul_color[0], ul_color[1], ul_color[2], ul_color[3],
			ur_color[0], ur_color[1], ur_color[2], ur_color[3])
		self.set_this(x)
	def __repr__(self): return "Rectangle "+Stretchable.__repr__(self)
	def set_width(self, width):
		oldsize = self.get_size()
		Stretchable.set_width(self, width)
		self.this = _pointless.nodefactory_rectangle(width, oldsize.y.max, 
			self.ll_color[0],self.ll_color[1],self.ll_color[2],self.ll_color[3],
			self.lr_color[0],self.lr_color[1],self.lr_color[2],self.lr_color[3],
			self.ul_color[0],self.ul_color[1],self.ul_color[2],self.ul_color[3],
			self.ur_color[0],self.ur_color[1],self.ur_color[2],self.ur_color[3])

## FIXME Lout operates with 6 gap modes!!!
eGlueType = Enumeration("eGlueType",
	[ "EDGE_TO_EDGE",
	"REF_TO_REF",
	"REF_TO_REF_FORCE"
	]
)

class Penalty(Node):
	def __init__(self, penalty=0):
		Node.__init__(self, this=None, context=None)
		self.discardable = 1
		self.penalty = penalty

class Glue(Stretchable):
	""" 
		DESCRIPTION:
			This is the base class for all glue elements both
			horizontal, vertical and absolute.
	"""
	def __init__(self, context, natural_width=0, stretchability=0, 
			shrinkability=0, height=0, natural_height=0, 
			type = eGlueType.EDGE_TO_EDGE, stretch_dim=0, 
			pos=v3(0,0,0)):
		Stretchable.__init__(self, natural_width=natural_width, 
			stretchability=stretchability, shrinkability=shrinkability, 
			natural_height=natural_height, context=context, 
			stretch_dim=stretch_dim)
		self.type = type
		self.pos = pos
		self.discardable = 1
	def __repr__(self):
		if self.is_vertical(): 
			return ("VGlue (%f, %f)" % (self._size.y.max, self.get_natural().y))
		elif self.is_absolute(): 
			return ("AGlue (%f, %f, %f)" % (self.pos.x, self.pos.y, self.pos.z))
		else: 
			return ("HGlue (%f, %f)" % (self._size.x.max, self.get_natural().x))
	def is_vertical(self):
		return self.stretch_dim == 1
	def is_horizontal(self):
		return self.stretch_dim == 0
	def is_absolute(self):
		return (self.pos.x != 0) or (self.pos.y != 0) or (self.pos.z != 0)
	def set_width(self, width):
		assert(not self.is_vertical())
		self._size.x.max = width
	def merge(self, glue):
		if glue.is_absolute():
			self.set_natural(glue)
		else:
			assert(self.stretch_dim == glue.stretch_dim) 
			self.set_natural(self.get_natural() + glue.get_natural())
	def zero(self):
		self.set_natural(v3(0,0,0))

class Group(Node):
	""" 
		DESCRIPTION: 
			FIXME
		METHODS
			add_and_pos(self, node, pos):
				adds a node at a specific (local) position and expands 
				the _size accordingly. Furthermore new hrefs and/or 
				vrefs are added as needed.
			add_right(self, node, space):
				FIXME
			add_below(self, node, space, type):
				FIXME
			add_absolute(self, node, pos):
	"""
	def __init__(self, this):
		Node.__init__(self, this, context=None)
		self.children = []
	def __repr__(self):
		return "<C Group instance at %s>" % (self.this,)
	def add(self, arg):
		self.children.append(arg)
		if arg.this:
			return _pointless.group_add(self.this, arg.this)
		else: return 1
	def add_pos_by_ref(self, self_ref, node, node_ref, pos):
		node_pos = self_ref + pos - node_ref
		self.add_and_pos(node, node_pos)
	def add_and_pos(self, node, pos):
		# Expand size
		self._size.x.max = max(self._size.x.max, node._size.x.max + pos.x)
		self._size.x.min = min(self._size.x.min, node._size.x.min + pos.x)
		self._size.y.max = max(self._size.y.max, node._size.y.max + pos.y)
		self._size.y.min = min(self._size.y.min, node._size.y.min + pos.y)
		self._size.z.max = max(self._size.z.max, node._size.z.max + pos.z)
		self._size.z.max = min(self._size.z.min, node._size.z.min + pos.z)
		# Append / move refs
		if node._hrefs[-1].x + pos.x > self._hrefs[-1].x:
			self._hrefs.append(copy.copy(node._hrefs[-1]))
			self._hrefs[-1] += v3(pos.x, 0, 0)
		if node._hrefs[0].x + pos.x < self._hrefs[0].x:
			self._hrefs = [copy.copy(node._hrefs[0])] + self._hrefs
			self._hrefs[0] += v3(pos.x, 0, 0)
		if node._vrefs[-1].y + pos.y < self._vrefs[-1].y:
			self._vrefs.append(copy.copy(node._vrefs[-1]))
			self._vrefs[-1] += v3(0, pos.y, 0)
		if node._vrefs[0].y + pos.y > self._vrefs[0].y:
			self._vrefs = [copy.copy(node._vrefs[0])] + self._vrefs
			self._vrefs[0] += v3(0, pos.y, 0)
		# Place node
		node.set_pos(pos)
		self.add(node)
	def add_right(self, node, space):
		#if len(self.children) == 0:
			#space = 0
		pos = copy.copy(self._hrefs[-1])
		pos.x = self._size.x.max
		self.add_pos_by_ref(pos, node, node._hrefs[0], v3(space,0,0))
		#print "Node pos after: ", node.get_pos()
	def add_below(self, node, space, type):
		if type == eGlueType.EDGE_TO_EDGE:
			pos = copy.copy(self._vrefs[-1])
			pos.y = self._size.y.min - node._size.y.max - space
			self.add_and_pos(node, pos)
		elif type == eGlueType.REF_TO_REF:
			if len(self.children) == 0: space = 0
			space = max(space, self._vrefs[-1].y - self._size.y.min + 
				node._size.y.max - node._vrefs[0].y)
			self.add_pos_by_ref(self._vrefs[-1], node, node._vrefs[0], 
				v3(0,-space,0))
		elif type == eGlueType.REF_TO_REF_FORCE:
			self.add_pos_by_ref(self._vrefs[-1], node, node._vrefs[0], 
				v3(0,-space,0))
		else:
			assert(0) # invalid type
	def add_absolute(self, node, pos):
		node.set_pos(pos)
		self.add(node)
	def _debug_draw(self):
		size = self.get_size()
		r = random()*0.9 + 0.1
		g = random()*0.9 + 0.1
		b = random()*0.9 + 0.1
		a = 0.2
		print size, self.get_pos()
		x = _pointless.nodefactory_rectangle(
			size.x.size(), size.y.size(),
			r, g, b, a,
			r, g, b, a,
			r, g, b, a,
			r, g, b, a,
			)
		self.debug_gfx = Leaf(x)
		print self.debug_gfx.get_size()
		self.debug_gfx.set_pos(v3(size.x.min,size.y.min,0))
		#renderer.get_root().add(self.debug_gfx)
		self.add(self.debug_gfx)
	def init_debug_draw(self):
		for c in self.children: c.init_debug_draw()
		if self.debug_draw: self._debug_draw()

class Step:
	"""Token to signify a new slide_step"""
	def __init__(self, controller_token):
		assert(isinstance(controller_token, ControllerToken))
		self.controller_token = controller_token
	def __repr__(self): return "Step(%s)" % self.controller_token.__repr__()

class ControllerToken:
	"""Token to signify a controller"""
	def __init__(self, controller_class, parameters, slide_no, slide_step):
		self.controller_class = controller_class
		self.parameters       = parameters
		self.slide_no         = slide_no
		self.slide_step       = slide_step
	def __repr__(self): return "ControllerToken(%s)"%str(self.controller_class)
	def produce(self):
		c = self.controller_class(self.parameters)
		c.slide_no = self.slide_no
		c.slide_step = self.slide_step
		return c

# --------------------------------------------------------------------------- #
# paragraph breaking

class Cell:
	""" 
		DESCRIPTION:
			Used to mark out each =cel cmd, cf. in pll.rowlist
	"""
	def __repr__(self):
		return "Cell"

class NodeList(list):
	""" 
		DESCRIPTION:
			The base class for lists of tokens (inner nodes). The elements in 
			the list are positioned next to each other according to the glue. 
			If the glue between two elements is missing then glue of size 0 is 
			assumed. 
		METHODS:
		set_step_span
			This is a util function helping to determine how far down the 
			tree the controllers needs to be attached. We want the controllers 
			to be as high as possible (i.e. only _one_ controller fading in 
			the slide, rather than n controllers on each leaf fading in the 
			slide). Look for the mark_slidesteps() and attach_controllers() 
			functions elsewhere. In effect each inner node can cover a 'span' 
			of slide-steps. If an inner-node covers a span of the type [n,m] 
			where m>n, we need to look further down. If an inner node covers 
			a span of the type [n,n] the controller for step n can be attached 
			to the node.
		arrange
			This base class will simply make a "pass" in the arrange method 
			but derived classes are supposed to set glue in one of the three 
			dimensions. 
		makenode
			A method that is supposed to return a Group node containing 
			the content of the NodeList, properly formatted. Note that
			the list of tokens contains quite different types of
			tokes, e.g. Controller, Leaf and Glue tokens. All
			Controller objects are attached to the the Group node that
			this method eventually returns.
	"""
	def __init__(self, nodes):
		self += nodes
		self.step_span = None
		self.group = None
	def __repr__(self):
		return " Span(%s)%s" % (self.step_span.__repr__(), list.__repr__(self))
	def set_step_span(self, span):
		self.step_span = span
	def arrange(self, available):
		pass
	def makenode(self, available):
		self.arrange(available)
		x = _pointless.nodefactory_group("Group")
		if not x: return None
		g = Group(x)
		controllers = []
		glue = None
		for n in self:
			if isinstance(n, ControllerToken):
				## simply collect all controllers in the list of tokens
				controllers.append(n)
				continue
			if isinstance(n, Glue):
				if not glue:
					glue = n
				elif (glue.is_vertical() and n.is_absolute()):
					## HACK: this is a meaningless case that we wont allow
					assert(0), "You can't have vglue before absolute glue"
				else:
					glue.merge(n)
				continue
			if isinstance(n, list):
				assert(isinstance(n, NodeList)),"Only NodeList (and derived) allowed here"
				w = g.get_size().x.size()
				if glue:
					if glue.is_vertical(): w = 0
					else: w += glue.get_size().x.size()
				n.makenode(available - v3(w,0,0))
				n = n.group
			assert(isinstance(n, Node)), \
				"How did this end up in the list of tokens: %s. Context: %s." \
				% (str(n), str(self))
			if glue == None:
				## no glue implies horizontal glue of size 0
				g.add_right(n, 0)
			elif glue.is_vertical():
				g.add_below(n, glue.get_size().y.size(), glue.type)
			elif glue.is_absolute():
				g.add_absolute(n, glue.pos)
			else:
				g.add_right(n, glue.get_size().x.size())
			glue = None
		for c in controllers:
			## attach controllers to the group node
			controller = c.produce()
			controller.set_node(g)
			pll.slideshow.add_observer(controller)
		###g.debug_draw = 1
		self.group = g
		#return g

def strip_discardable_front(l):
	while len(l) and isinstance(l[0], Node) and l[0].discardable:
		del(l[0])
def strip_discardable_back(l):
	while len(l) and isinstance(l[-1], Node) and l[-1].discardable:
		l.pop()

class CellList(NodeList):
	"""
	"""
	def __init__(self, nodes):
		NodeList.__init__(self, nodes)

class RowList(NodeList):
	""" 
		DESCRIPTION: 
			A RowList is a list of table rows. Thus, the RowList will contain
			CellLists (the rows), and each CellList must contain only NodeLists
			(the cells of the row).
			The makenode() method is overridden in order to layout all the
			cells first, and then modify the width of the individual cells
			according to the widest cell in each column.
	"""
	def __init__(self, nodes, num_cells, cellpadding, rowpadding):
		NodeList.__init__(self, nodes)
		for row in nodes:
			if type(row) != CellList: continue
			for e in row:
				assert(type(e) == NodeList)
		self.num_cells = num_cells
		self.set = 0
		self.cellpadding = cellpadding
		self.rowpadding = rowpadding
	def __repr__(self):
		return "\nRowList"+NodeList.__repr__(self)
	def makenode(self, available):
		newrowlist = []
		numcols = 0
		for row in self:
			if not type(row) == CellList:
				newrowlist.append(row)
				continue
			if len(row) > numcols:
				numcols = len(row)
			newrow = []
			for cell in row:
				cell.makenode(available)
				newrow.append(cell.group)
			row[:] = newrow
			newrowlist.append(row)
		self[:] = newrowlist
		colwidths = [0.0 for x in range(0,numcols)]
		# Find colwidths
		for row in self:
			if not isinstance(row, CellList): continue
			for cell, col_num in zip(row, range(len(row))):
				if cell.get_size().x.size() > colwidths[col_num]:
					colwidths[col_num] = cell.get_size().x.size()
		# Adjust cellsizes
		for row in self:
			if not isinstance(row, CellList): continue
			for cell, col_num in zip(row, range(len(row))):
				space = colwidths[col_num] - cell.get_size().x.size()
				assert(space >= 0.0)
				cell._size.x.max += space
				cell._size.x.max += self.cellpadding
				if col_num == 0:
					cell._size.x.min -= self.cellpadding
					cell._hrefs[0].x -= self.cellpadding
		avail = sum(colwidths) + self.cellpadding * (len(colwidths)+1)
		NodeList.makenode(self, v3(avail, 0, 0))
				
class StretchList(NodeList):
	""" 
		DESCRIPTION:
			This class overrides the arrange() method in order to set
			the glue. That is, we call the set_stretch() method on
			all the Glue tokens in the list with proper arguments.
		METHODS:
		set_glues
			Compute the width of the elements in the list of tokens
			and compare it with the desired width. Compute a stretch
			factor based on the difference between these two numbers
			and apply this on all the glue elements in the list. The
			list of tokens will then end up having the desired width.
	"""
	def __init__(self, nodes):
		NodeList.__init__(self, nodes)
		self.set = 0
	def __repr__(self):
		return "\nStretchList"+NodeList.__repr__(self)
	def arrange(self, available):
		assert(not self.set), "Oops. arrange called twice on StretchList"
		##### self.set_glues(available.x)
		self.set = 1
		wanted_width = available.x
		natural_width = sum(map(lambda x: x._size.x.min + x._size.x.max, self))
		if natural_width < wanted_width:
			stretch_wanted = wanted_width - natural_width
			glues = filter(lambda x: isinstance(x, Stretchable), self)
			stretch_available = sum(map(lambda x: x.stretchability, glues))
			pllog.debug(gt("Stretching (%f -> %f)"%(natural_width, wanted_width)))
			if stretch_available == 0: return
			## FIXME MAGIC - why 1.0 ???
			stretch_factor = 1.0*stretch_wanted / stretch_available 
			pllog.debug(gt("want %f, natural %f, avail %f" % 
				(wanted_width, natural_width, stretch_available)))
			for g in glues:
				g.set_stretch(stretch_factor)
		elif natural_width > wanted_width:
			shrink_wanted = natural_width - wanted_width
			glues = filter(lambda x: isinstance(x, Stretchable), self)
			shrink_available = sum(map(lambda x: x.shrinkability, glues))
			if shrink_available == 0: return
			## FIXME MAGIC - why 1.0 ???
			shrink_factor = 1.0*shrink_wanted / shrink_available 
			for g in glues:
				g.set_stretch(-shrink_factor)
	
class ParList(NodeList):
	""" 
		DESCRIPTION:
			This class overrides the arrange method in order to do line 
			breaking. The new self.list contains elements whose
			width doesn't exceed the available width.
		METHODS:
		arrange(self, available):
			The self.list is traversed and rebuild using two temporary lists 
			(lines and controllers), one consisting of all the controller 
			elements in the self.list and another consisting of line elements, 
			i.e. list of tokens constituting lines. 
	"""
	def __init__(self, nodes, adjustment = eAdjustmentType.ADJUSTED):
		NodeList.__init__(self, nodes)
		self.enumtype = adjustment
	def __repr__(self):
		return "\nParList"+NodeList.__repr__(self)
	def arrange(self, available):
		lines = []	
		controllers = []
		width = 0
		# pointer to the last processed Glue element and its width
		last_breakable = None
		## flag telling whether we should make a break 
		break_now = 0
		line = []
		for n in self:
			if isinstance(n, ControllerToken):
				controllers.append(n)
				continue
			if isinstance(n, NodeList):
				## group the elements into a single new element
				n.makenode(available - v3(width, 0, 0))
				n = n.group
				assert(n)
			if isinstance(n, Glue):
				if n.is_vertical():
					lines.append(self.make_line(line))
					lines.append(n)
					line  = []
					width = 0
					continue
				last_breakable = (n, width)
			if isinstance(n, Penalty) and n.penalty == -INFINITY:
				break_now = 1
			line.append(n)
			width += n.get_size().x.size()
			if width > available.x or break_now:
				break_now = 0
				if not last_breakable:
					pllog.warn(gt("Unable to break line"))
					if len(lines): lines.append(pll.vertical_bl_glue(
						n.context.font.baselineskip))
					lines.append(self.make_line(line))
					width = 0
					line  = []
				else:
					where = line.index(last_breakable[0])
					baselineskip = last_breakable[0].context.font.baselineskip
					if len(lines):
						lines.append(pll.vertical_bl_glue(baselineskip))
					lines.append(self.make_line(line[:where]))
					width -= last_breakable[1]
					line  = line[where+1:]
				last_breakable = None
		if len(line):
			## proper handling of the last elements 
			if len(lines) and not isinstance(lines[-1], Glue):
				lines.append(pll.vertical_bl_glue(
					line[-1].context.font.baselineskip))
			lines.append(self.make_line(line, last_line=1))
		self[:] = controllers + lines
	def make_line(self, line, last_line=0):
		prefix = []
		suffix = []
		if self.enumtype == eAdjustmentType.CENTERED:
			strip_discardable_front(line)
			strip_discardable_back(line)
			prefix = [pll.hfill_glue()]
			suffix = [pll.hfill_glue()]
		elif last_line:
			strip_discardable_back(line)
			suffix = [pll.hfill_glue()]
		return StretchList(prefix+line+suffix)

# --------------------------------------------------------------------------- #
# The Renderer

class Renderer:
	""" 
		DESCRIPTION: 
			This is essentaillty the python wrapping of the renderer
			which is implemented in the CPP world.
	"""
	def __init__(self):
		_pointless.renderer_init(
			plsetup.geometry.get_width(),
			plsetup.geometry.get_height(),
			plsetup.geometry.get_placement()
		)
		self.root = Group(_pointless.renderer_get_root())
		self.debug_osd = 0
		self.resolutions = str(_pointless.renderer_get_resolutions()).split(' ')
		if '' in self.resolutions: 
			self.resolutions.remove('') 
		self.resolutions.sort(lambda x,y: int.__cmp__(int(x.split('x')[0]),\
			int(y.split('x')[0])))
		self.resolution_index = 0
		myindex = _pointless.renderer_get_resolution()
		if myindex:
			self.resolution_index = \
				self.resolutions.index(_pointless.renderer_get_resolution())
		else:
			self.resolution_index = 0
		self.fullscreen = 0
	def __del__(self, avoidexception=_pointless.renderer_shutdown):
		avoidexception()
	def get_root(self):
		return self.root
	def get_width(self):
		return int(_pointless.renderer_get_width())
	def get_height(self):
		return int(_pointless.renderer_get_height())
	def get_fullscreen(self):
		return _pointless.renderer_get_fullscreen()
	def toggle_fullscreen(self):
		_pointless.renderer_toggle_fullscreen();
		if self.get_fullscreen():
			self.fullscreen = 1
		else:
			self.fullscreen = 0
	def set_resolution(self, res):
		if self.resolutions.__contains__(res):
			_pointless.renderer_set_resolution(res)
			return 1
		else:
			return 0
	def increase_resolution(self):
		self.resolution_index += 1
		if self.resolution_index >= len(self.resolutions):
			self.resolution_index=0
		self.set_resolution(self.resolutions[self.resolution_index])
	def decrease_resolution(self):
		self.resolution_index -= 1
		if self.resolution_index < 0:
			self.resolution_index = len(self.resolutions) - 1
		self.set_resolution(self.resolutions[self.resolution_index])
	def get_resolution(self):
		return str(_pointless.renderer_get_resolution())
	def get_resolutions(self):
		return self.resolutions
	def print_at(self,str,x=2,y= 1):
		## default debug location in the python part of pointless is (2,1)
		if self.debug_osd: _pointless.renderer_print_at(x, y, str)
	def getevent(self, block):
		return _pointless.renderer_getevent(block)
	def flip_debug_osd(self):
		self.debug_osd = not self.debug_osd
		_pointless.renderer_set_debug_osd(self.debug_osd)
	def set_wireframe(self, wf):
		_pointless.renderer_set_wireframe(wf)
	def get_wireframe(self):
		return _pointless.renderer_get_wireframe()
	def frame(self):
		return _pointless.renderer_frame()
	def load_textures(self):
		_pointless.renderer_load_textures()
	def init_gl(self):
		_pointless.renderer_init_gl()
	def rootadd(self, arg):
		self.root.add(arg)
	def newroot(self):
		_pointless.renderer_newroot()
		self.root = Group(_pointless.renderer_get_root())
	def make_screenshot(self, filename=None, scale=1.0):
		if not filename:
			used_nums = ['000'] + [
				x[-7:-4] for x in os.listdir(".")
				if re.search("^plshot[0-9]{3}\.png$", x)
			]
			used_nums.sort()
			filename = "plshot%03i.png" % (int(used_nums[-1]) + 1)
		if not _pointless.renderer_make_screenshot(filename, scale):
			pllog.abort(gt("Writing screenshot to '%s' failed"%filename))
		else:
			pllog.info(gt("Wrote screenshot in '%s'"%filename))
		
# --------------------------------------------------------------------------- #
# Director of slideshow

class SlideInfo:
	def __init__(self, slide, steps):
		self.slide = slide
		self.steps = steps

class SlideShow:
	""" 
		DESCRIPTION: 
			This class is used to control the slideshow, the methods
			describes the actions to take when one switches from one
			state to another.
	"""
	def __init__(self):
		self.current_slide = None
		self.slide_no = 0
		self.slide_step = 0
		self.slides = {}
		self.observers = {}
		self.lo_fps = 0
	def set_lo_fps(self, lo_fps):
		self.lo_fps = lo_fps
	def add_slide(self, slide, slide_no, steps):
		self.slides[slide_no] = SlideInfo(slide, steps)
	def add_observer(self, observer):
		slide_no = observer.slide_no
		slide_step = observer.slide_step
		key = "%i:%i" % (slide_no, slide_step)
		if not self.observers.has_key(key):
			self.observers[key] = []
		self.observers[key].append(observer)
	def get_observers(self, slide_no, slide_step):
		key = "%i:%i" % (slide_no, slide_step)
		if self.observers.has_key(key): return self.observers[key]
		return []
	def cur_slide(self):
		return self.slides[self.slide_no]
	def get_slide(self, slide_no):
		if self.slides.has_key(slide_no): return self.slides[slide_no]
		else: return None
	def next_step(self):
		new_slide_step = self.slide_step + 1
		if new_slide_step > self.cur_slide().steps:
			self.next_slide()
		else:
			self.enter_step(new_slide_step)
			self.slide_step = new_slide_step
	def prev_step(self):
		new_slide_step = self.slide_step - 1
		if new_slide_step < 0:
			self.prev_slide(at_bottom=1)
		else:
			self.leave_step(self.slide_step)
			self.slide_step = new_slide_step
	def next_slide(self):
		new_slide_no   = self.slide_no + 1
		new_slide_step = 0
		if not self.get_slide(new_slide_no):
			return
		self.move_to_slide(new_slide_no)
		self.move_to_step (new_slide_step, instant=1)
	def prev_slide(self, at_bottom = 0):
		new_slide_no  = self.slide_no - 1
		if not self.get_slide(new_slide_no):
			return
		if at_bottom: new_slide_step = self.get_slide(new_slide_no).steps
		else: new_slide_step = 0
		self.move_to_slide(new_slide_no)
		self.move_to_step (new_slide_step, instant=1)
	def move_to_slide(self, slide_no, instant=0):
		if self.slide_no == slide_no: return
		self.leave_slide(self.slide_no, instant)
		self.enter_slide(slide_no, instant)
		self.slide_no = slide_no
	def enter_step(self, step_no, instant=0):
		instant = instant or self.lo_fps
		for o in self.get_observers(self.slide_no, step_no):
			o.on_enter(instant=instant)
	def leave_step(self, step_no, instant=0):
		instant = instant or self.lo_fps
		for o in self.get_observers(self.slide_no, step_no):
			o.on_leave(instant=instant)
	def enter_slide(self, slide_no, instant=0):
		instant = instant or self.lo_fps
		for o in self.get_observers(slide_no, 0):
			o.on_enter(instant=instant)
	def leave_slide(self, slide_no, instant=0):
		instant = instant or self.lo_fps
		for o in self.get_observers(slide_no, 0):
			o.on_leave(instant=instant)
	def move_to_step(self, slide_step, instant=0):
		instant = instant or self.lo_fps
		for s in range(1, self.cur_slide().steps + 1):
			if s <= slide_step:
				self.enter_step(s, instant=instant)
			else:
				self.leave_step(s, instant=instant)
		self.slide_step = slide_step
	def first_slide(self):
		self.move_to_slide(1)
		self.move_to_step(0, instant=1)
	def last_slide(self):
		self.move_to_slide(max(self.slides.keys()))
		self.move_to_step(0, instant=1)
	def set_show_pos(self, slide_no, slide_step):
		self.move_to_slide(slide_no)
		self.move_to_step(slide_step, instant=1)
	def get_show_pos(self):
		return (self.slide_no, self.slide_step)
	def __arrange(self):
		assert(0)
		for slide_info in self.slides.values():
			slide = slide_info.slide
			slide.arrange()
	def status_str(self):
		return "Slide: %i:%i" % self.get_show_pos()
	def reset(self):
		for k in self.observers.keys():
			for o in self.observers[k]:
				o.init()
				o.on_leave(instant=1)

# --------------------------------------------------------------------------- #
		
class SlideStepping:
	def __init__(self, slide_no, slide_steps):
		self.slide_no = slide_no
		self.slide_steps = slide_steps

class Pll:
	""" 
		DESCRIPTION: FIXME
	"""
	def __init__(self, slideshow, scale, slide_size):
		self.slideshow = slideshow
		self.commands = {}
		self.line_no = 0
		self.largefactor = 2
		self.hugefactor = 3
		self.smallfactor = 0.75
		self.tinyfactor = 0.5
		self.basesize = PlSize(str(ploptions.default_fontsize))
		self.fs = FontStateStack();
		self.fs.set_font(ploptions.default_fontname, 
			self.basesize.pixels(scale))
		self.fs.push()
		self.scale = scale
		self.slide_size = slide_size
		self.background = self.group("Background")
		renderer.rootadd(self.background)
		self.slide_no = 0
		self.slide_step = 0
		self.slide_steps = []
		self.no_slide_step = 0
		self.filename = None
		self.tempfileno = 0
		self.tempfiles = []
	def create_cache(self):
		assert(self.filename), "Unable to make cachefiles before setting filename"
		if not os.path.isdir(self.cachedir):
			try:
				os.mkdir(self.cachedir)
			except OSError:
				hashname = md5.new(self.cachedir).hexdigest()
				hashname = hashname[:16]
				self.cachedir = "/tmp/"+hashname ## FIXME hardcoding
				if not os.path.isdir(self.cachedir):
					try:
						os.mkdir(self.cachedir)
					except OSError:
						raise RuntimeError, gt("Unable to create cache")
	def get_tempfilename(self):
		self.create_cache()
		self.tempfileno += 1
		tempfilename = "%s/pl%03i-tmp" % (self.cachedir, self.tempfileno)
		self.tempfiles.append(tempfilename)
		return tempfilename
	def get_cachefilename(self, realfilename):
		self.create_cache()
		realfilename = os.path.abspath(realfilename)
		hashname = md5.new(realfilename).hexdigest()
		hashname = hashname[:16]
		return "%s/%s"%(self.cachedir, hashname)
	def set_lines(self, lines, filename):
		if filename:
			self.filename = filename
			self.cachedir = "%s.cache" % os.path.abspath(filename)
		self.lines = lines
		self.line_no = 1
	def new_slide(self):
		self.slide_no += 1
		self.slide_step = 0
		stepping = SlideStepping(self.slide_no, 0)
		self.slide_steps.append(stepping)
	def next_slide_step(self):
		self.slide_step += 1
		self.slide_steps[-1].slide_steps = self.slide_step
	def skip_next_step(self):
		self.no_slide_step = 1
	def get_slide_size(self):
		return self.slide_size
	def import_module(self, modulename):
		try:
			module = __import__(modulename)
		except ImportError, e:
			raise PllParseError(e)
		module.pll = self
		module.register()
		for cmd in vars(module).keys():
			if not isinstance(vars(module)[cmd], Command): continue
			self.commands[cmd] = vars(module)[cmd]
	def add_command(self, name, execute, slidestep = 0, transform = 0, 
			break_condition = re_newline, vars = 0):
		cmd = Command(name, execute, slidestep, transform, break_condition, vars)
		self.commands[cmd.name] = cmd
	def include_file(self, filename):
		## FIXME what happens if f includes filename f
		filename = plfiles.find_file(filename)
		file  = plfiles.open_file(filename)
		lines = file.readlines()
		self.lines = lines + self.lines
	def eval(self, text):
		return self.parse([text], break_condition=re_eof, skip_white=1)
	def parse(self, lines, break_condition=re_newline, remove_break=0, skip_white=0):
		result = NodeList([])
		while len(lines):
			# Check for break-condition
			if break_condition:
				r = break_condition.search(lines[0])
				if r:
					if remove_break:
						lines[0] = lines[0][r.end(0):]
					break
			# End of line
			if lines[0] == "\n" or lines[0] == "":
				lines[:] = lines[1:]
				if len(lines): lines[0] = " " + lines[0]
				self.line_no += 1
				continue
			# Check for comment
			r = re_comment.search(lines[0])
			if r:
				lines[0] = lines[0][r.end():]
				continue
			# Check for =command
			r = re_command.search(lines[0])
			if r:
				cmd_name = r.group(1)
				if cmd_name == 'slide' and len(result):
					break
				cmd_method = r.group(2)
				#cmd_args = r.group(4)
				cmd_break_condition = None
				cmd_remove_break = 0
				if re_begin.search(cmd_name):
					cmd_name = cmd_name[6:] # chop off 'begin-'
					cmd_break_condition = re.compile('^=end-'+cmd_name)
					cmd_remove_break = 1
				if self.commands.has_key(cmd_name):
					cmd = self.commands[cmd_name]
					lines[0] = lines[0][r.end(0):]
					cmd_args = None
					if r.group(3):
						in_quote = None
						for c, i in zip(lines[0], range(len(lines[0]))):
							if in_quote:
								if c == in_quote:
									in_quote = None
								continue
							if c in "'\"":
								in_quote = c
								continue
							if c == ')': break ## FIXME what if in a .set cmd
						cmd_args = lines[0][0:i]
						lines[0] = lines[0][i+1:]
					if not cmd_break_condition:
						cmd_break_condition = cmd.break_condition
					cmd_result = None
					if cmd_method == '.set':
						try:
							## FIXME the regex stuff is a HACK to allow 
							## specifications s.a. 0.0, 40.0, -400.0
							r = re.search(',', cmd_args)
							if r:
								a = re.sub(':',':(',cmd_args)
								cmd_args = a+')'
							h = eval('{'+cmd_args+'}')
						except:
							h = {}
						cmd.set(h)
						continue
					if cmd_args:
						args = cmd.parse_args(cmd_args)
					else:
						args = []
					if cmd.transform:
						text = ""
						while len(lines):
							r = cmd_break_condition.search(lines[0])
							if not r:
								text += lines[0]
								lines[:] = lines[1:]
								self.line_no += 1
							else:
								text += lines[0][:r.start(0)]
								if cmd_remove_break:
									lines[0] = lines[0][r.end(0):]
								else:
									lines[0] = lines[0][r.start(0):]
								break
						try:
							lines[0] = (cmd.execute(args, text, vars=None) +
								lines[0])
						except PllCommandError, e:
							print gt("Line %i: Error executing =%s." %
								(self.line_no, cmd_name))
							for eline in str(e).split("\n"):
								print "**", eline
						continue
					if cmd.slidestep:
						if self.no_slide_step:
							self.no_slide_step = 0
						else:
							self.next_slide_step()
							result += [self.item_step()]
					parser = lambda : self.parse(lines, break_condition = 
						cmd_break_condition, remove_break = cmd_remove_break)
					try:
						cmd_result = cmd.execute(args = args, parser = parser,\
							vars = Scope(cmd_name, vs))
					except PllCommandError, e:
						print gt("Line %i: Error executing =%s." %
							(self.line_no, cmd_name))
						for eline in str(e).split("\n"):
							print "**", eline
						continue
					assert(type(cmd_result) == list), "Commands must return lists"
					result += cmd_result
					continue
			# Linewrap
			r = re_linewrap.search(lines[0])
			if r:
				lines[:] = lines[1:]
				continue
			# Escapes
			r = re_escapes.search(lines[0])
			if r:
				lines[0] = lines[0][r.end():]
				if r.group(0) == ' ':
					result += [self.interword_glue()]
				else:
					result += [self.letter(ord(r.group(1)))]
				continue
			# Blocks
			r = re_blockstart.search(lines[0])
			if r:
				self.fs.push()
				vs.push()
				lines[0] = lines[0][1:]
				continue
			r = re_blockend.search(lines[0])
			if r:
				self.fs.pop()
				vs.pop()
				lines[0] = lines[0][1:]
				continue
			# Letters
			r = re_letters.search(lines[0])
			if r:
				lines[0] = lines[0][r.end():]
				result += [self.letter(ord(r.group(0)))]
				continue
			# Space
			r = re_space.search(lines[0])
			if r:
				lines[0] = lines[0][r.end():]
				if not skip_white and len(result) and \
						not isinstance(result[-1], Glue):
					result += [self.interword_glue()]
				continue
			badinput = lines[0][0:10]
			badinput = badinput.replace('\n','\\n')
			errmsg = gt("Line %i: Parse error: Don't know how to handle: \
				'%s'... ")%(self.line_no, badinput)
			lines[:]=[]
			raise PllParseError(errmsg)
		strip_discardable_front(result)
		strip_discardable_back(result)
		return result
	def letter(self, unicode):
		fs = self.fs.current()
		x = _pointless.nodefactory_letter(fs.font.this, unicode,
			fs.color[0], fs.color[1], fs.color[2], fs.color[3])
		if x: return Leaf(x, context=fs)
		else:
			pllog.warn(gt("No glyph for unicode %i in font '%s'" % \
				(unicode,fs.font.fontname)))
			pllog.warn(gt("inserting a '#' to substitute unicode %i" % unicode))
			x = _pointless.nodefactory_letter(fs.font.this, 0x0023,
				fs.color[0], fs.color[1], fs.color[2], fs.color[3])
			if x: return Leaf(x, context=fs)
			else: assert(0), "Unable to find char '#' in this font. Giving up"
	def linebreak(self):
		return Penalty(penalty=-INFINITY)#, context=self.fs.current())
	def interword_glue(self, scale=1.0):
		fs = self.fs.current()
		glue = Glue(
			## FIXME MAGIC 
			natural_width = fs.font.size * 0.3333333 * scale,
			stretchability = 5,
			context = self.fs.current()
		)
		return glue
	def hfill_glue(self, skip=0):
		glue = Glue(
			natural_width = skip,
			stretchability=INFINITY,
			context = self.fs.current()
		)
		return glue
	def absolute_glue(self, abs_pos=v3(0,40,20)):
		glue = Glue(
			context = self.fs.current(), 
			pos = abs_pos 
		)
		return glue
	def vertical_glue(self, skip):
		glue = Glue(
			natural_height = skip,
			type = eGlueType.EDGE_TO_EDGE,
			context = self.fs.current(),
			stretch_dim = 1
		)
		return glue
	def vertical_bl_glue(self, skip, force=0):
		if force:
			gluetype = eGlueType.REF_TO_REF_FORCE
		else:
			gluetype = eGlueType.REF_TO_REF
		glue = Glue(natural_height = skip,
			type = gluetype,
			context = self.fs.current(),
			stretch_dim = 1)
		return glue
	def backgroundimage(self, stretch, filename):
		f = self.find_file(filename)
		if f:
			x = _pointless.nodefactory_backgroundimage(
				renderer.get_width(), renderer.get_height(), stretch, f)
			if x: return Leaf(x)
	def image(self, filename, center):
		gray_alpha = 0
		f = self.find_file(filename)
		if f:
			x = _pointless.nodefactory_image(f, center, gray_alpha)
			if x: return Leaf(x)
	def gaimage(self, filename, center):
		gray_alpha = 1
		f = self.find_file(filename)
		if f:
			x = _pointless.nodefactory_image(f, center, gray_alpha)
			if x: return Leaf(x)
	def rectangle(self, width, height, ll_color, lr_color=-1, 
			ul_color=-1, ur_color=-1):
		if lr_color == -1:
			lr_color = ll_color
		if ul_color == -1:
			ul_color = ll_color
		if ur_color == -1:
			ur_color = ll_color
		glue = Rectangle(
			natural=v3(width, height, 0),
			context=self.fs.current(),
			ll_color=ll_color,
			lr_color=lr_color,
			ul_color=ul_color,
			ur_color=ur_color
			)
		return glue
	def group(self, name):
		x = _pointless.nodefactory_group(name)
		if x: return Group(x)
	def paragraphlist(self, nodes, adjustment = eAdjustmentType.ADJUSTED):
		return ParList(nodes, adjustment = adjustment)
	def nodelist(self, nodes):
		return NodeList(nodes)
	def celllist(self, nodes):
		cells = [NodeList([])]
		for n in nodes:
			if isinstance(n, Cell):
				# each =cel cmd implies a glue element and a new Nodelist
				# cells[-1].append(self.hfill_glue(20))  FIXME
				cells.append(NodeList([]))
				continue
			cells[-1].append(n)
		for c in cells:
			strip_discardable_front(c)
			strip_discardable_back(c)
		return CellList(cells)
	def rowlist(self, nodes, cellpadding, rowpadding):
		#nodes = [ x for x in nodes if isinstance(x, list) ]
		nodes = [ x for x in nodes
			if not (isinstance(x, Glue) and x.is_horizontal()) ]
		return RowList(nodes, len(nodes), cellpadding, rowpadding)
	def cell(self):
		return Cell()
	def stretchlist(self, nodes):
		return StretchList(nodes)
	def slide_controller(self, parameters = 0):
		if not parameters:
			## proper default parameters:
			parameters = {'in_delay':0.8, 'move_in':(0.0, 0.0, 0.0), 
				'move_out':(0.0, 0.0,0.0)} 
		return ControllerToken(AlphaFader, parameters, self.slide_no, self.slide_step)
	def item_controller(self, slide_step=None):
		if not slide_step:
			slide_step = self.slide_step
		return ControllerToken(AlphaFader, {}, self.slide_no, slide_step)
	def item_step(self):
		return Step(self.item_controller())
	def find_file(self, filename):
		return plfiles.find_file(filename)

# --------------------------------------------------------------------------- #
# Font stuff

class Font:
	def __init__(self, fontname, size):
		f = plfiles.find_file(fontname)
		self.fontname = fontname
		self.size = size
		self.this = _pointless.get_font(f, size)
		if not self.this:
			raise RuntimeError, ("Could not load font '%s'" % fontname)
		metric = _pointless.get_font_metric(self.this)
		self.baselineskip = metric["baselineskip"]
		self.underline_position = metric["underline_position"]
		self.underline_thickness = metric["underline_thickness"]
		if abs(self.underline_position) > self.size:
			pllog.warn(gt("Possibly broken ttf metrics. Underline pos: %f" % 
				self.underline_position))
			self.underline_position = 0
		if abs(self.underline_thickness) > self.size:
			pllog.warn(gt("Possibly broken ttf metrics. Underline thickness: %f" %
				self.underline_thickness))
			self.underline_thickness = 0

class FontState:
	def __init__(self, font, color):
		self.font  = font
		self.color = color
	def copy(self):
		return FontState(self.font, self.color)

class FontStateStack:
	def __init__(self):
		self.fontcache = {}
		self.stack = []
	def _get_font(self, fontname, size):
		fontkey = ("%s:%i" % (fontname, size))
		if self.fontcache.has_key(fontkey):
			return self.fontcache[fontkey]
		else:
			font = None
			try:
				font = Font(fontname, size)
			except (RuntimeError), e:
				pllog.warn(gt("%s. Substituting with default font" % e))
				try:
					font = Font(ploptions.default_fontname, size)
				except (RuntimeError), e:
					pllog.abort(gt("Could not load default font: %s") % e)
			self.fontcache[fontkey] = font
			return font
	def push(self):
		if len(self.stack) == 0:
			pllog.abort(gt("No default font set"));
		# FIXME: Cache the fontcontexts
		self.stack.append(self.stack[-1].copy())
	def pop(self):
		self.stack.pop()
	def current(self):
		return self.stack[-1]
	def set_size(self, size):
		if not len(self.stack):
			pllog.abort(gt("Set size but no font selected"))
		font = self._get_font(self.current().font.fontname, size)
		# FIXME: Cache the fontcontexts
		self.stack[-1] = self.stack[-1].copy()
		self.stack[-1].font = font
	def set_font(self, fontname, size=None):
		assert(fontname), "Must specify name of font"
		if not size:
			assert(len(self.stack))
			size=self.current().font.size
		font = self._get_font(fontname, size)
		if not len(self.stack):
			fontstate = FontState(font, [1, 1, 1, 1])
			self.stack.append(fontstate)
		else:
			# FIXME: Cache the fontcontexts
			self.stack[-1] = self.stack[-1].copy()
			self.stack[-1].font = font

# --------------------------------------------------------------------------- #
# auxilary functions used by main 

def sum(list):
	return reduce(lambda x,y: x+y, list, 0)

def print_tree(root, indent = 0):
	pllog.warn("%s(%i)%s" % (" "*indent, sys.getrefcount(root), root.name()))
	if isinstance(root, Group):
		for c in root.children:
			print_tree(c, indent+1)

def mark_slidesteps(slide, step, collector):
	if isinstance(slide, list):
		assert(isinstance(slide, NodeList))
		start = step
		for x in slide:
			step = mark_slidesteps(x, step, collector)
			if isinstance(x, Step):
				slide.remove(x)
		slide.set_step_span([start, step])
		return step
	else:
		if isinstance(slide, Step):
			step += 1
			collector[step] = slide
		return step

def attach_controllers(slide, collector):
	if isinstance(slide, list):
		assert(isinstance(slide, NodeList))
		if slide.step_span[0] == slide.step_span[1]:
			step = slide.step_span[0]
			if step > 0:
				assert(collector.has_key(step))
				#print "Attaching ", collector[step].controller_token, "to ", slide
				slide.append(collector[step].controller_token)
		else:
			for x in slide:
				attach_controllers(x, collector)
	else:
		if isinstance(slide, Leaf):
			pllog.warn(gt("Internal error. Slide step inside group (on %s)."%slide))

def setup_controllers(slide):
	collector = {}
	mark_slidesteps(slide, 0, collector)
	attach_controllers(slide, collector)

def run_slideshow():
	slideshow   = SlideShow()
	if ploptions.powersave:
		slideshow.set_lo_fps(1)
	#print_tree(renderer.get_root()) ## solely for debugging purposes
	if ploptions.banner:
		filedata = [x+'\n' for x in ploptions.banner.split('\n')]
	else:
		f = plfiles.find_file(ploptions.inputfile)
		file = plfiles.open_file(f)
		filedata = file.readlines()
	## Calculate size of slide based on window size and aspect ratio.
	corrected_height = renderer.get_height()*ploptions.aspect
	corrected_width  = renderer.get_width()
	corrected_size   = min(corrected_height, corrected_width)
	real_height = corrected_size / ploptions.aspect
	real_width  = corrected_size
	left_offset = (renderer.get_width() - real_width) / 2.0 ## FIXME MAGIC
	left_margin = 0.03 * real_width ## FIXME MAGIC
	top_offset  = (renderer.get_height() - real_height) / 2.0 ## FIXME MAGIC
	top_margin  = 0.03 * real_height ## FIXME MAGIC
	slide_size = v3(
		size(left_offset, left_offset + real_width),
		size(top_offset,  top_offset  + real_height),
		size(0,0)
	)
	global pll
	scale = real_height / plsetup.geometry.get_height()
	pll = Pll(slideshow, scale, slide_size)
	pll.import_module("pllbasics")
	pll.set_lines(filedata, ploptions.inputfile)
	slidelist = []
	while len(pll.lines):
		sys.stdout.write("[")
		sys.stdout.flush()
		slide = pll.parse(pll.lines, break_condition=None, skip_white=1)
		if slide: slidelist.append(slide)
		else: pllog.warn(gt("Internal problem: parser returned empty slide"))
		sys.stdout.write("%i] " % len(slidelist))
		sys.stdout.flush()
	print
	for slide in slidelist:
		setup_controllers(slide)
	no_of_slides = 0
	for slide, stepping in map(None, slidelist, pll.slide_steps):
		## FIXME MAGIC
		available = v3(real_width - 2.0*left_margin, INFINITY, INFINITY)
		if DEBUG:
			print "================================="
			print "Slide: ", stepping.slide_no
			print "Dump : ", slide
			print "================================="
		slide.makenode(available)
		g = slide.group
		g.set_pos(v3(left_margin + left_offset, top_offset + real_height - 
			(top_margin + g._size.y.max - g._vrefs[0].y), 0)) 
		g.init_debug_draw()
		renderer.rootadd(g)
		slideshow.add_slide(g, stepping.slide_no, stepping.slide_steps)
		no_of_slides += 1
	renderer.load_textures()
	renderer.init_gl()
	slideshow.reset()
	if ploptions.export:
		## dump screenshots and export dumpinfo
		i = 0
		name = plsetup.merger.get_exportbasename()
		thumb_width = 200 ## FIXME MAGIC
		thumb_scale = thumb_width / float(renderer.get_width())
		thumb_height = renderer.get_height() * thumb_scale
		while((i < no_of_slides)):
			i += 1
			slideshow.move_to_slide(i, instant=1)
			# Hack - something larger than the number of steps
			slideshow.move_to_step(100000, instant=1);  ## FIXME MAGIC
			running_no = plsetup.merger.next_running_no()
			shot_name = name + running_no + ".png"
			thumb_shot_name = "small-"+shot_name 
			renderer.frame()
			renderer.make_screenshot(plsetup.merger.get_exportdir() + shot_name)
			renderer.make_screenshot(plsetup.merger.get_exportdir() + \
					thumb_shot_name, thumb_scale)
			plsetup.merger.add_merge_info([running_no, \
			                       shot_name, \
								   renderer.get_width(), \
								   renderer.get_height(), \
								   thumb_shot_name, \
								   thumb_width, \
								   thumb_height])
		if ploptions.exportraw:
			plsetup.merger.dump_raw()
		else:
			plsetup.merger.merge()
		return 0
	else:
		if PROFILE: return
		## Normal slideshow behavior 
		slideshow.set_show_pos(1,0)
		## now, call frame and wait for key-events, the real main-loop
		while(1):
			renderer.frame()
			renderer.print_at(slideshow.status_str())
			event = renderer.getevent(ploptions.powersave)
			if not event: continue
			(event_type, event_data) = event
			# Unknown event
			if event_type != _pointless.KEYDOWN and \
			   event_type != _pointless.MOUSEBUTTONDOWN and \
			   event_type != _pointless.VIDEORESIZE:
				pllog.warn(gt("Unknown event type"));
				continue
			# Resize event
			if event_type == _pointless.VIDEORESIZE:
				# Hack to get rid of extra resize event from wm's running
				# in opaque resize mode...
				for a in range(1,100):
					e = renderer.getevent(0)
				pllog.debug(gt("Resizing to %d x %d" % (event_data)));
				renderer.newroot()
				return 1
			# Mouse button pressed
			if event_type == _pointless.MOUSEBUTTONDOWN:
				but = event_data[0]
				if but in (_pointless.BUTTON_LEFT, 
						_pointless.BUTTON_WHEEL_DOWN):
					slideshow.next_step()
				elif but in (_pointless.BUTTON_RIGHT, 
						_pointless.BUTTON_WHEEL_UP):
					slideshow.prev_step()
				elif but == _pointless.BUTTON_MIDDLE:
					pass
				else:
					pllog.warn(gt("Unknown mouse button: %i"%but))
				continue
			# Key pressed
			(sym, mod, type) = event_data
			if sym == _pointless.K_q:
				return 0
			elif sym == _pointless.K_d:
				renderer.flip_debug_osd()
			elif sym == _pointless.K_w:
				renderer.set_wireframe(not renderer.get_wireframe())
			elif (sym == _pointless.K_SPACE) or \
				(sym == _pointless.K_DOWN) or \
				(sym == _pointless.K_j):
				slideshow.next_step()
			elif (sym == _pointless.K_UP) or (sym == _pointless.K_k):
				slideshow.prev_step()
			elif (sym == _pointless.K_RIGHT) or (sym == _pointless.K_l):
				slideshow.next_slide()
			elif (sym == _pointless.K_LEFT) or (sym == _pointless.K_h):
				slideshow.prev_slide()
			elif (sym == _pointless.K_0):
				slideshow.first_slide()
			elif (sym == _pointless.K_g) and (mod == _pointless.KMOD_SHIFT):
				slideshow.last_slide()
			elif (sym == _pointless.K_s):
				renderer.make_screenshot(scale=0.5)
			elif (sym == _pointless.K_i):
				print gt("Current geometry: %sx%s") % \
				(plsetup.geometry.get_width(), plsetup.geometry.get_height())
				print gt("Current resolution: '%s'") % renderer.get_resolution()
				print gt("List of possible resolutions:")
				l=renderer.get_resolutions()
				print  "\n".join(l)
			elif (sym == _pointless.K_PLUS):
				## FIXME, international keyboards, DK keyboard ('+'='shift+0')
				renderer.increase_resolution()
			elif (sym == _pointless.K_MINUS):
				renderer.decrease_resolution()
			elif (sym == _pointless.K_RETURN) and (mod == _pointless.KMOD_CTRL):
				renderer.toggle_fullscreen()
	
def init_renderer():
	"""Initialize renderer"""
	global renderer
	renderer = Renderer()
	if ploptions.print_resolutions:
		l=renderer.get_resolutions()
		if l:
			print gt("List of possible resolutions:")
			current = renderer.get_resolution()
			for r in l:
				if r == current: print " *", 
				else: print "  ", 
				print  r
		sys.exit()
	if ploptions.wanted_resolution:
		if not renderer.set_resolution(ploptions.wanted_resolution):
			print gt("Resolution '%s' not supported.") % \
				ploptions.wanted_resolution;
			sys.exit()
	if ploptions.fullscreen:
		pllog.info(gt("Start in fullscreen mode"))
		renderer.toggle_fullscreen();

# --------------------------------------------------------------------------- #
# Main loop

def main():
	plsetup.process_args()
	pllog.debug(gt("CVS id for this particular version %s") % __id__)
	init_renderer()
	## Run slideshow. Returns true on resize etc, and false when exit
	## has been requested
	try:
		if PROFILE:
			profile.run("pointless.run_slideshow()", "pointless.prof")
			s = pstats.Stats('pointless.prof')
			s.sort_stats('cumulative')
			s.print_stats()
		else:
			while(run_slideshow()): pass
		return 0
	except (PllParseError), e:
		print e
		return 1

# --------------------------------------------------------------------------- #

